class MonkeyBananaProblem:
    def __init__(self):
        
        self.initial_state = ("door", "floor", "window", False)
        self.goal_state = ("center", "on_block", "center", True)

    def is_goal(self, state):
        """Check if the goal state is reached."""
        return state == self.goal_state

    def get_successors(self, state):
        """Generate possible next states from the current state."""
        monkey, vertical, block, has_banana = state
        successors = []
        
        
        if vertical == "floor":
            for new_pos in ["door", "window", "center"]:
                if new_pos != monkey:
                    successors.append(((new_pos, vertical, block, has_banana), f"Monkey walks to {new_pos}"))
        
        
        if vertical == "floor" and monkey == block:
            for new_pos in ["door", "window", "center"]:
                if new_pos != block:
                    successors.append(((new_pos, vertical, new_pos, has_banana), f"Monkey pushes block to {new_pos}"))
        
        
        if vertical == "floor" and monkey == block:
            successors.append(((monkey, "on_block", block, has_banana), "Monkey climbs on the block"))
        
       
        if vertical == "on_block" and block == "center":
            successors.append(((monkey, vertical, block, True), "Monkey grabs the banana"))
        
        return successors

    def solve(self):
        """Perform DFS to find the sequence of moves leading to the goal."""
        stack = [(self.initial_state, [])] 
        visited = set()
        
        while stack:
            current_state, path = stack.pop()
            
            if self.is_goal(current_state):
                return path  
            
            if current_state not in visited:
                visited.add(current_state)
                
                for next_state, action in self.get_successors(current_state):
                    stack.append((next_state, path + [action]))
        
        return None 


problem = MonkeyBananaProblem()
solution = problem.solve()

if solution:
    print("Solution found:")
    for step in solution:
        print(f"- {step}")
else:
    print("No solution found.")



import heapq

def a_star_search(graph, start, goal, heuristics):
    open_list = [] 
    heapq.heappush(open_list, (0 + heuristics[start], start, [start], 0))
    
    visited = set()

    while open_list:
        f_cost, current, path, g_cost = heapq.heappop(open_list)

        if current in visited:
            continue
        
        visited.add(current)

        if current == goal:
            print("Path found:", ' -> '.join(path))
            print("Total cost:", g_cost)
            return

        for neighbor, cost in graph[current]:
            if neighbor not in visited:
                new_g = g_cost + cost
                new_f = new_g + heuristics[neighbor]
                heapq.heappush(open_list, (new_f, neighbor, path + [neighbor], new_g))

    print("No path found.")

graph = {
    'A': [('B', 1), ('C', 4)],
    'B': [('D', 2), ('E', 5)],
    'C': [('F', 3)],
    'D': [('G', 1)],
    'E': [('G', 2)],
    'F': [('G', 5)],
    'G': []
}

heuristics = {
    'A': 7,
    'B': 6,
    'C': 5,
    'D': 4,
    'E': 3,
    'F': 2,
    'G': 0
}

a_star_search(graph, 'A', 'G', heuristics)




from collections import deque

def bfs(graph, start_node):
    visited = set()
    queue = deque([start_node])
    visited.add(start_node)

    while queue:
        current_node = queue.popleft()
        print(current_node, end=" ")

        for neighbour in graph[current_node]:
            if neighbour not in visited:
                visited.add(neighbour)
                queue.append(neighbour)

graph = {
    'A': ['B', 'C'],
    'B': ['A', 'D', 'E'],
    'C': ['A', 'F'],
    'D': ['B'],
    'E': ['B', 'F'],
    'F': ['C', 'E']
}

bfs(graph, 'A')



def dfs(graph, node, visited=None):
    if visited is None:
        visited = set()
    
    visited.add(node)
    
    print(node, end=" ")
    
    for neighbor in graph[node]:
        if neighbor not in visited:
            dfs(graph, neighbor, visited)

graph = {
    'A': ['B', 'C'],
    'B': ['A', 'D', 'E'],
    'C': ['A', 'F'],
    'D': ['B'],
    'E': ['B', 'F'],
    'F': ['C', 'E']
}

dfs(graph, 'A')



from setuptools import setup, find_packages

setup(
    name='pypract',
    version='0.1.0',
    description='A collection of practical codes for learning',
    packages=find_packages(),  
    install_requires=[],  
)



def isValid(board, row, col, N):
    # Check the row to the left of current column
    for i in range(col):
        if board[row][i] == 1:
            return False
    
    # Check the upper diagonal to the left
    for i, j in zip(range(row, -1, -1), range(col, -1, -1)):
        if board[i][j] == 1:
            return False
    
    # Check the lower diagonal to the left
    for i, j in zip(range(row, N, 1), range(col, -1, -1)):
        if board[i][j] == 1:
            return False
    
    return True

def solveNQueen(board, col, N):
    if col >= N:
        return True
    

    for i in range(N):
        if isValid(board, i, col, N):
            board[i][col] = 1
            if solveNQueen(board, col + 1, N):
                return True
            # Backtrack
            board[i][col] = 0
    
    return False

def printBoard(board, N):
    for i in range(N):
        for j in range(N):
            print("Q" if board[i][j] == 1 else "_", end=" ")
        print()

def nQueens(N):
    board = [[0] * N for _ in range(N)] 
    if not solveNQueen(board, 0, N):
        print("No solution exists")
    else:
        printBoard(board, N)

N = 8  
nQueens(N)
